from __future__ import annotations
import colorama
from pathlib import Path
import threading
from typing import List, Optional, Tuple, TYPE_CHECKING
from . import util


if TYPE_CHECKING:
    from .util import PathOrStr
    from .actions import Action


class Task:
    """
    Task to be executed.
    """
    def __init__(
            self,
            name: str,
            *,
            dependencies: Optional[List["PathOrStr"]] = None,
            targets: Optional[List["PathOrStr"]] = None,
            action: Optional[Action] = None,
            task_dependencies: Optional[List[Task]] = None,
            location: Optional[Tuple[str, int]] = None,
            ) -> None:
        self.name = name
        self.dependencies = dependencies or []
        self.targets = [Path(path) for path in (targets or [])]
        self.action = action
        self.task_dependencies = task_dependencies or []
        self.location = location or util.get_location()

    def execute(self, stop: Optional[threading.Event] = None) -> None:
        if self.action:
            self.action.execute(self, stop)

    def __hash__(self) -> int:
        return hash(self.name)

    def format(self, color: str = None) -> str:
        name = self.name
        if color:
            name = f"{color}{name}{colorama.Fore.RESET}"
        filename, lineno = self.location
        return f"<task `{name}` @ {filename}:{lineno}>"

    def __repr__(self) -> str:
        return self.format()



from cook import Manager, Task
from cook.actions import CompositeAction, FunctionAction, ModuleAction, SubprocessAction
from cook.contexts import Context, create_target_directories, FunctionContext, create_group, \
    normalize_action, normalize_dependencies
from cook.controller import Controller
from pathlib import Path
import pytest
import sqlite3
import sys
from typing import List


def test_function_context(m: Manager) -> None:
    tasks: List[Task] = []

    def func(t: Task) -> Task:
        tasks.append(t)
        return t

    with FunctionContext(func):
        m.create_task("my-task")
    m.create_task("my-other-task")

    task, = tasks
    assert task.name == "my-task"


def test_missing_task_context(m: Manager) -> None:
    with pytest.raises(ValueError, match="did not return a task"), FunctionContext(lambda _: None):
        m.create_task("my-task")


def test_context_management(m: Manager) -> None:
    with pytest.raises(RuntimeError, match="no active contexts"), Context():
        m.contexts = []
    with pytest.raises(RuntimeError, match="unexpected context"), Context():
        m.contexts.append("something else")


def test_create_target_directories(m: Manager, tmp_wd: Path, conn: sqlite3.Connection) -> None:
    filename = tmp_wd / "this/is/a/hierarchy.txt"
    with normalize_action(), create_target_directories():
        task = m.create_task("foo", targets=[filename], action=["touch", filename])
    assert not filename.parent.is_dir()

    controller = Controller(m.resolve_dependencies(), conn)
    controller.execute(task)
    assert filename.parent.is_dir()


def test_create_target_directories_with_multiple_targets(m: Manager, tmp_wd: Path, conn: sqlite3) \
        -> None:
    filenames = [tmp_wd / "this/is/a/hierarchy.txt", tmp_wd / "this/is/a/hierarchy2.txt"]
    with normalize_action(), create_target_directories():
        for filename in filenames:
            task = m.create_task(filename.name, targets=[filename], action=["touch", filename])
    assert not filename.parent.is_dir()

    controller = Controller(m.resolve_dependencies(), conn)
    controller.execute(task)
    assert filename.parent.is_dir()


def test_normalize_action(m: Manager) -> None:
    with normalize_action():
        task = m.create_task("foo", action="bar")
        assert isinstance(task.action, SubprocessAction) and task.action.args[0] == "bar" \
            and task.action.kwargs["shell"]

        task = m.create_task("bar", action=["baz"])
        assert isinstance(task.action, SubprocessAction) and task.action.args[0] == ["baz"] \
            and not task.action.kwargs.get("shell")

        actions = [SubprocessAction("hello", shell=True), SubprocessAction("world")]
        task = m.create_task("baz", action=actions)
        assert isinstance(task.action, CompositeAction) and task.action.actions == tuple(actions)

        task = m.create_task("xyz", action=lambda x: None)
        assert isinstance(task.action, FunctionAction)

        task = m.create_task("fizz", action=[pytest, "foo", "bar"])
        assert isinstance(task.action, ModuleAction) \
            and task.action.args[0] == [sys.executable, "-m", "pytest", "foo", "bar"]

        with pytest.raises(ValueError, match="must not be an empty list"):
            m.create_task("buzz", action=[])


def test_group_no_tasks(m: Manager) -> None:
    with pytest.raises(RuntimeError, match="no tasks"), create_group("g"):
        pass


def test_group(m: Manager) -> None:
    with create_group("g"):
        t1 = m.create_task("t1")
        t2 = m.create_task("t2")
    assert m.tasks["g"].task_dependencies == [t1, t2]


def test_normalize_dependencies(m: Manager) -> None:
    with create_group("g") as g:
        base = m.create_task("base")
    with normalize_dependencies():
        task = m.create_task("task1", dependencies=[g])
        assert task.task_dependencies == [g.task]

        task = m.create_task("task2", dependencies=[base])
        assert task.task_dependencies == [base]

        task = m.create_task("task3", task_dependencies=["g"])
        assert task.task_dependencies == [g.task]



from cook import Manager
from cook.contexts import create_group, normalize_dependencies
from cook.util import CookError
import logging
from pathlib import Path
import pytest


def test_resolve_dependencies(m: Manager) -> None:
    Path("input1.txt").write_text("input1.txt")
    Path("input2.txt").write_text("input2.txt")

    with normalize_dependencies():
        intermediate = m.create_task("intermediate", dependencies=["input1.txt", "input2.txt"],
                                     targets=["intermediate.txt"])
        with create_group("outputs") as outputs:
            output1 = m.create_task("output1", dependencies=["intermediate.txt"],
                                    targets=["output1.txt"])
            output2 = m.create_task("output2", targets=["output2.txt"],
                                    dependencies=["intermediate.txt", "input2.txt", "output1.txt"])
        special = m.create_task("special", dependencies=["intermediate.txt"])
        dependent = m.create_task("dependent", task_dependencies=[output1])
    dependencies = m.resolve_dependencies()

    expected = {
        output1: {intermediate},
        output2: {intermediate, output1},
        special: {intermediate},
        dependent: {output1},
        outputs.task: {output1, output2},
    }
    for task, deps in expected.items():
        assert set(dependencies.successors(task)) == deps


def test_missing_file(m: Manager) -> None:
    with normalize_dependencies():
        m.create_task("has_missing_file_dependency", dependencies=["missing-file.txt"])
    with pytest.raises(FileNotFoundError, match="does not exist nor is"):
        m.resolve_dependencies()


def test_conflicting_targets(m: Manager) -> None:
    m.create_task("foo", targets=["bar"])
    m.create_task("baz", targets=["bar"])
    with pytest.raises(ValueError, match="both have target"):
        m.resolve_dependencies()


def test_symlink_targets(m: Manager, caplog: pytest.LogCaptureFixture, tmp_wd: Path) -> None:
    file2 = Path("file2")
    file1 = Path("file1")
    m.create_task("task1", targets=[file1])
    m.create_task("task2", targets=[file2])
    file2.symlink_to("file3")

    with caplog.at_level(logging.WARNING):
        m.resolve_dependencies()
    assert "is a symlink" in caplog.messages[0]

    file2.unlink()
    file2.symlink_to(file1)
    with pytest.raises(ValueError, match="both have target"):
        m.resolve_dependencies()


def test_same_name(m: Manager) -> None:
    m.create_task("foo")
    with pytest.raises(ValueError, match="task with name 'foo' already exists"):
        m.create_task("foo")


def test_get_manager_instance() -> None:
    with pytest.raises(ValueError, match="no manager"):
        Manager.get_instance()
    with Manager() as m:
        assert Manager.get_instance() is m
    with pytest.raises(RuntimeError, match="unexpected manager"), Manager():
        Manager._INSTANCE = "asdf"
    with pytest.raises(ValueError, match="already active"), Manager(), Manager():
        pass


def test_dependency_graph(m: Manager) -> None:
    with normalize_dependencies():
        bar = m.create_task("bar", targets=["a"])
        foo = m.create_task("foo", dependencies=["a"])
    dependencies = m.resolve_dependencies()
    assert set(dependencies.successors(foo)) == {bar}
    assert set(dependencies.predecessors(bar)) == {foo}


def test_dependency_graph_cycle(m: Manager) -> None:
    with normalize_dependencies():
        m.create_task("bar", targets=["bar"], dependencies=["foo"])
        m.create_task("foo", targets=["foo"], dependencies=["bar"])
    with pytest.raises(CookError, match="contains a cycle"):
        m.resolve_dependencies()



import argparse
import colorama
from datetime import datetime
import fnmatch
import importlib.util
import logging
import os
from pathlib import Path
import re
import sqlite3
import sys
import textwrap
from typing import Iterable, List, Optional

from .contexts import create_target_directories, normalize_action, normalize_dependencies
from .controller import Controller, QUERIES
from .manager import Manager
from .task import Task
from .util import FailedTaskError, format_datetime, format_timedelta


LOGGER = logging.getLogger("cook")


class NoMatchingTaskError(ValueError):
    def __init__(self, patterns: Iterable[re.Pattern], hidden_tasks_available: bool):
        self.hidden_tasks_available = hidden_tasks_available
        patterns = [f"`{pattern}`" for pattern in patterns]
        if len(patterns) == 1:
            message = f"found no tasks matching pattern {patterns[0]}"
        else:
            *patterns, last = patterns
            message = "found no tasks matching patterns " + ", ".join(patterns) \
                + (", or " if len(patterns) > 1 else " or ") + last
        if hidden_tasks_available:
            message = message + "; use --all or -a to include hidden tasks"
        super().__init__(message)


class Args:
    tasks: Iterable[re.Pattern]
    re: bool
    all: bool


class Command:
    """
    Abstract base class for commands.
    """
    NAME: Optional[str] = None
    ALLOW_EMPTY_PATTERN: bool = False

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--re", "-r", action="store_true",
                            help="use regular expressions for pattern matching instead of glob")
        parser.add_argument("--all", "-a", action="store_true",
                            help="include tasks starting with `_` prefix")
        parser.add_argument("tasks", nargs="*" if self.ALLOW_EMPTY_PATTERN else "+",
                            help="task or tasks to execute as regular expressions or glob patterns")

    def execute(self, controller: Controller, args: argparse.Namespace) -> None:
        raise NotImplementedError

    def discover_tasks(self, controller: Controller, args: Args) -> List[Task]:
        task: Task
        tasks: List[Task] = []

        # Get tasks based on the pattern matching.
        if not args.tasks:
            tasks = list(controller.dependencies)
        else:
            for task in controller.dependencies:
                if args.re:
                    match = any(re.match(pattern, task.name) for pattern in args.tasks)
                else:
                    match = any(fnmatch.fnmatch(task.name, pattern) for pattern in args.tasks)
                if match:
                    tasks.append(task)

        # Store whether any of the candidates are hidden by default.
        has_hidden_task = any(task.name.startswith("_") for task in tasks)

        # Filter out hidden tasks if desired unless the name is an exact match to a specified
        # pattern.
        if not args.all:
            tasks = [task for task in tasks if not task.name.startswith("_") or task.name in
                     args.tasks]

        if not tasks:
            raise NoMatchingTaskError(args.tasks, not args.all and has_hidden_task)
        return tasks


class ExecArgs(Args):
    jobs: int


class ExecCommand(Command):
    """
    Execute one or more tasks.
    """
    NAME = "exec"

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--jobs", "-j", help="number of concurrent jobs", type=int, default=1)
        super().configure_parser(parser)

    def execute(self, controller: Controller, args: ExecArgs) -> None:
        tasks = self.discover_tasks(controller, args)
        controller.execute(tasks, num_concurrent=args.jobs)


class LsArgs(Args):
    stale: bool
    current: bool


class LsCommand(Command):
    """
    List tasks.
    """
    NAME = "ls"
    ALLOW_EMPTY_PATTERN = True

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--stale", "-s", action="store_true", help="only show stale tasks")
        parser.add_argument("--current", "-c", action="store_true", help="only show current tasks")
        super().configure_parser(parser)

    def execute(self, controller: Controller, args: LsArgs) -> None:
        tasks = self.discover_tasks(controller, args)
        tasks = [task.format(colorama.Fore.YELLOW if is_stale else colorama.Fore.GREEN)
                 for is_stale, task in zip(controller.is_stale(tasks), tasks)]
        print("\n".join(tasks))

    def discover_tasks(self, controller: Controller, args: LsArgs) -> List[Task]:
        if args.current and args.stale:
            raise ValueError("only one of `--stale` and `--current` may be given at the same time")
        tasks = super().discover_tasks(controller, args)
        if args.stale:
            return [task for stale, task in zip(controller.is_stale(tasks), tasks) if stale]
        elif args.current:
            return [task for stale, task in zip(controller.is_stale(tasks), tasks) if not stale]
        return tasks


class InfoCommand(LsCommand):
    """
    Display information about one or more tasks.
    """
    NAME = "info"

    def execute(self, controller: Controller, args: LsArgs) -> None:
        tasks = self.discover_tasks(controller, args)
        stales = controller.is_stale(tasks)

        stale_string = f"{colorama.Fore.YELLOW}stale{colorama.Fore.RESET}"
        current_string = f"{colorama.Fore.GREEN}current{colorama.Fore.RESET}"
        indent = "    "

        task: Task
        for stale, task in zip(stales, tasks):
            # Show the status.
            parts = [
                f"status: {stale_string if stale else current_string}",
            ]
            # Show when the task last completed and failed.
            last = controller.connection.execute(
                "SELECT last_completed, last_failed FROM tasks WHERE name = :name",
                {"name": task.name}
            ).fetchone() or (None, None)
            for key, value in zip(["completed", "failed"], last):
                if value is None:
                    parts.append(f"last {key}: -")
                    continue
                parts.append(f"last {key}: {format_timedelta(datetime.now() - value)} ago "
                             f"({format_datetime(value)})")
            # Show dependencies and targets.
            task_dependencies = list(sorted(controller.dependencies.successors(task),
                                            key=lambda t: t.name))
            task_dependencies = [
                dep.format(colorama.Fore.YELLOW if is_stale else colorama.Fore.GREEN)
                for is_stale, dep in zip(controller.is_stale(task_dependencies), task_dependencies)
            ]
            items = [
                ("dependencies", task.dependencies),
                ("targets", task.targets),
                ("task_dependencies", task_dependencies),
            ]
            for key, value in items:
                value = "\n".join(map(str, value))
                if value:
                    parts.append(f"{key}:\n{textwrap.indent(value, indent)}")
                else:
                    parts.append(f"{key}: -")
            parts.append(f"action: {task.action if task.action else '-'}")

            parts = textwrap.indent('\n'.join(parts), indent)
            print(f"{task}\n{parts}")


class ResetArgs(argparse.Namespace):
    tasks: Iterable[re.Pattern]
    re: bool


class ResetCommand(Command):
    """
    Reset the status of one or more tasks.
    """
    NAME = "reset"

    def execute(self, controller: Controller, args: ResetArgs) -> None:
        tasks = self.discover_tasks(controller, args)
        controller.reset(*tasks)


class Formatter(logging.Formatter):
    COLOR_BY_LEVEL = {
        "DEBUG": colorama.Fore.MAGENTA,
        "INFO": colorama.Fore.BLUE,
        "WARNING": colorama.Fore.YELLOW,
        "ERROR": colorama.Fore.RED,
        "CRITICAL": colorama.Fore.WHITE + colorama.Back.RED,
    }
    RESET = colorama.Fore.RESET + colorama.Back.RESET

    def format(self, record: logging.LogRecord) -> str:
        color = self.COLOR_BY_LEVEL[record.levelname]
        formatted = f"{color}{record.levelname}{self.RESET}: {record.getMessage()}"
        if record.exc_info:
            formatted = f"{formatted}\n{self.formatException(record.exc_info)}"
        return formatted


def __main__(cli_args: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser("cook")
    parser.add_argument("--recipe", help="file containing declarative recipe for tasks",
                        default="recipe.py", type=Path)
    parser.add_argument("--module", "-m", help="module containing declarative recipe for tasks")
    parser.add_argument("--db", help="database for keeping track of assets", default=".cook")
    parser.add_argument("--log-level", help="log level", default="info",
                        choices={"error", "warning", "info", "debug"})
    subparsers = parser.add_subparsers()
    subparsers.required = True

    for command_cls in [ExecCommand, LsCommand, InfoCommand, ResetCommand]:
        subparser = subparsers.add_parser(command_cls.NAME, help=command_cls.__doc__)
        command = command_cls()
        command.configure_parser(subparser)
        subparser.set_defaults(command=command)

    args = parser.parse_args(cli_args)

    handler = logging.StreamHandler()
    handler.setFormatter(Formatter())
    logging.basicConfig(level=args.log_level.upper(), handlers=[handler])

    with Manager() as manager:
        try:
            manager.contexts.extend([
                create_target_directories(),
                normalize_action(),
                normalize_dependencies(),
            ])
            if args.module:
                # Temporarily add the current working directory to the path.
                try:
                    sys.path.append(os.getcwd())
                    importlib.import_module(args.module)
                finally:
                    sys.path.pop()
            elif args.recipe.is_file():
                # Parse the recipe.
                spec = importlib.util.spec_from_file_location("recipe", args.recipe)
                recipe = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(recipe)
            else:  # pragma: no cover
                raise ValueError("recipe file or module must be specified; default recipe.py not "
                                 "found")
        except:  # noqa: E722
            LOGGER.fatal("failed to load recipe", exc_info=sys.exc_info())
            sys.exit(1)

    with sqlite3.connect(args.db, detect_types=sqlite3.PARSE_DECLTYPES) as connection:
        connection.executescript(QUERIES["schema"])
        controller = Controller(manager.resolve_dependencies(), connection)
        command: Command = args.command
        try:
            command.execute(controller, args)
        except KeyboardInterrupt:  # pragma: no cover
            LOGGER.warning("interrupted by user")
        except NoMatchingTaskError as ex:
            LOGGER.warning(ex)
            sys.exit(1)
        except FailedTaskError:
            sys.exit(1)


if __name__ == "__main__":
    __main__()



from datetime import datetime
import hashlib
import logging
import networkx as nx
from pathlib import Path
from queue import Empty, Queue
from sqlite3 import Connection
import sys
import threading
from typing import Dict, List, Optional, Sequence, Set, Tuple, TYPE_CHECKING, Union
from . import util

if TYPE_CHECKING:
    from .task import Task


LOGGER = logging.getLogger(__name__)
QUERIES = {
    "schema": """
    -- Information about the status of tasks.
    CREATE TABLE IF NOT EXISTS "tasks" (
        "name" TEXT PRIMARY KEY,
        "digest" TEXT NOT NULL,
        "last_completed" TIMESTAMP,
        "last_failed" TIMESTAMP
    );

    -- Information about files so we can cache digests.
    CREATE TABLE IF NOT EXISTS "files" (
        "name" TEXT PRIMARY KEY,
        "digest" TEXT NOT NULL,
        "last_digested" TIMESTAMP NOT NULL
    );
    """,
    "select_task": """
        SELECT "digest"
        FROM "files"
        WHERE "name" = :name
    """,
    "upsert_task_completed": """
        INSERT INTO "tasks" ("name", "digest", "last_completed")
        VALUES (:name, :digest, :last_completed)
        ON CONFLICT ("name") DO UPDATE SET "digest" = :digest, last_completed = :last_completed
    """,
    "upsert_task_failed": """
        INSERT INTO "tasks" ("name", "digest", "last_failed")
        VALUES (:name, '__failed__', :last_failed)
        ON CONFLICT ("name") DO UPDATE SET "digest" = '__failed__', last_failed = :last_failed
    """,
    "upsert_file": """
        INSERT INTO "files" ("name", "digest", "last_digested")
        VALUES (:name, :digest, :last_digested)
        ON CONFLICT ("name") DO UPDATE SET "digest" = :digest, last_digested = :last_digested
    """,
    "select_file": """
        SELECT "digest", "last_digested"
        FROM "files"
        WHERE "name" = :name AND last_digested > :last_modified
    """
}


class Controller:
    """
    Controller to manage dependencies and execute tasks.
    """
    def __init__(self, dependencies: nx.DiGraph, connection: Connection) -> None:
        self.dependencies = dependencies
        self.connection = connection
        self._digest_cache: Dict[Path, Tuple[float, bytes]] = {}

    def resolve_stale_tasks(self, tasks: Optional[List["Task"]] = None) -> Set["Task"]:
        self.is_stale(tasks or list(self.dependencies))
        return {node for node, data in self.dependencies.nodes(True) if data.get("is_stale")}

    def _evaluate_task_hexdigest(self, task: "Task") -> str:
        """
        Evaluate the digest of a task by combining the digest of all its dependencies.
        """
        dependencies = []
        for dependency in task.dependencies:
            dependency = Path(dependency).resolve()
            if not dependency.is_file():
                LOGGER.debug("dependency %s of %s is missing", dependency, task)
                return None
            dependencies.append(dependency)

        hasher = hashlib.sha1()
        for dependency in sorted(dependencies):
            hasher.update(bytearray.fromhex(self._evaluate_path_hexdigest(dependency)))

        # Add the hash of the action.
        if task.action and (hexdigest := task.action.hexdigest):
            hasher.update(bytearray.fromhex(hexdigest))
        return hasher.hexdigest()

    def _evaluate_path_hexdigest(self, path: Union[Path, str]) -> str:
        """
        Get the digest of a file.
        """
        # Try to return the cached digest.
        path = Path(path)
        stat = path.stat()
        name = str(path.resolve())
        params = {
            "name": name,
            "last_modified": datetime.fromtimestamp(stat.st_mtime)
        }
        digest = self.connection.execute(QUERIES["select_file"], params).fetchone()
        if digest:
            return digest[0]

        # Evaluate a new digest and cache it.
        digest = util.evaluate_hexdigest(path)
        params = {
            "name": name,
            "last_digested": datetime.now(),
            "digest": digest,
        }
        self.connection.execute(QUERIES["upsert_file"], params)
        self.connection.commit()
        return digest

    def is_stale(self, task: Union["Task", Sequence["Task"]]) -> Union[bool, List[bool]]:
        """
        Determine if one or more tasks are stale.

        Args:
            task: Task or tasks to check.

        Returns:
            If the task or tasks are stale.
        """
        if isinstance(task, Sequence):
            return [self.is_stale(x) for x in task]

        is_stale = self.dependencies.nodes[task].get("is_stale")
        if is_stale is not None:
            return is_stale
        is_stale = self._is_self_stale(task)
        successors = list(self.dependencies.successors(task))
        if successors:
            is_stale |= any(self.is_stale(successors))
        self.dependencies.nodes[task]["is_stale"] = is_stale
        return is_stale

    def _is_self_stale(self, task: "Task") -> bool:
        """
        Determine whether a task is *itself* stale irrespective of other tasks it may depend on.

        Args:
            task: Task to check.

        Returns:
            If the task is stale, ignoring dependencies.
        """
        # If there are no targets or the targets are missing, the task is stale.
        if not task.targets:
            LOGGER.debug('%s is stale because it has no targets', task)
            return True
        for target in task.targets:
            if not target.is_file():
                LOGGER.debug("%s is stale because its target `%s` is missing", task, target)
                return True

        # If there is no digest in the database, the task is stale.
        cached_digest = self.connection.execute("SELECT digest FROM tasks WHERE name = :name",
                                                {"name": task.name}).fetchone()
        if cached_digest is None:
            LOGGER.debug("%s is stale because it does not have a hash entry", task)
            return True

        # If one of the dependencies is missing, the task is stale.
        current_digest = self._evaluate_task_hexdigest(task)
        if current_digest is None:
            LOGGER.debug("%s is stale because one of its dependencies is missing", task)

        # If the digest has changed, the task is stale.
        cached_digest, = cached_digest
        if current_digest != cached_digest:
            LOGGER.debug("%s is stale because one of its dependencies has changed (cached digest: "
                         "%s, current digest: %s)", task, cached_digest, current_digest)
            return True

        LOGGER.debug("%s is up to date", task)
        return False

    def execute(self, tasks: "Task", num_concurrent: int = 1, interval: float = 1) -> None:
        """
        Execute one or more tasks.

        Args:
            tasks: Tasks to execute.
            num_concurrent: Number of concurrent threads to run.
        """
        if not isinstance(tasks, Sequence):
            tasks = [tasks]
        if not any(self.is_stale(tasks)):
            return

        # Start the worker threads.
        threads: List[threading.Thread] = []
        input_queue = Queue()
        output_queue = Queue()
        stop = util.StopEvent(interval)
        for i in range(num_concurrent):
            thread = threading.Thread(target=self._target, name=f"cook-thread-{i}",
                                      args=(stop, input_queue, output_queue), daemon=True)
            thread.start()
            threads.append(thread)

        # Get the subgraph of stale nodes.
        stale_nodes = [node for node, data in self.dependencies.nodes.data() if
                       data.get("is_stale")]
        dependencies: nx.DiGraph = self.dependencies.subgraph(stale_nodes).copy()

        # Initialize the input queue with leaf nodes.
        for node, out_degree in dependencies.out_degree():
            if out_degree == 0:
                input_queue.put((node, self._evaluate_task_hexdigest(node)))

        try:
            while dependencies.number_of_nodes():
                # Try to get the next item in the queue, continuing if there's nothing available.
                try:
                    item: Optional[Tuple["Task", sys._OptExcInfo]] = \
                        output_queue.get(timeout=interval)
                except Empty:  # pragma: no cover
                    continue

                assert item is not None, "output queue returned `None`; this is a bug"

                # Unpack the results.
                task, exc_info, digest = item
                if exc_info:
                    # Update the status in the database.
                    params = {
                        "name": task.name,
                        "last_failed": datetime.now(),
                    }
                    self.connection.execute(QUERIES["upsert_task_failed"], params)
                    self.connection.commit()
                    raise util.FailedTaskError(exc_info[1], task=task) from exc_info[1]
                else:
                    # Update the status in the database.
                    params = {
                        "name": task.name,
                        "digest": digest,
                        "last_completed": datetime.now(),
                    }
                    self.connection.execute(QUERIES["upsert_task_completed"], params)
                    self.connection.commit()

                # Check if the stop event is set and abort if so.
                if stop.is_set():
                    break

                # Add tasks that are now leaf nodes to the tree.
                predecessors = list(dependencies.predecessors(task))
                dependencies.remove_node(task)
                self.dependencies.add_node(task, is_stale=False)
                for node, out_degree in dependencies.out_degree(predecessors):
                    if out_degree == 0:
                        input_queue.put((node, self._evaluate_task_hexdigest(node)))
        finally:
            # Set the stop event and add "None" to the queue so the workers stop waiting.
            LOGGER.debug("set stop event for threads: %s", [thread.name for thread in threads])
            stop.set()
            for thread in threads:
                input_queue.put((None, None))

            # Shut down the worker threads.
            for thread in threads:
                thread.join(3 * stop.interval)
                if thread.is_alive():  # pragma: no cover
                    raise RuntimeError(f"thread {thread} failed to join")

    def _target(self, stop: util.StopEvent, input_queue: Queue, output_queue: Queue) -> None:
        LOGGER.debug(f"started thread `{threading.current_thread().name}`")
        while not stop.is_set():
            try:
                task: "Task"
                digest: str
                task, digest = input_queue.get(stop.interval)
            except Empty:  # pragma: no cover
                # It's unlikely there's nothing on the queue, but let's handle it anyway.
                continue
            # Check the stop event before executing the task; it may have been set while we were
            # waiting for the next task in the queue.
            if stop.is_set():
                break

            assert task is not None, "input queue returned `None`; this is a bug"

            start = datetime.now()
            try:
                # Execute the task.
                LOGGER.log(logging.DEBUG if task.name.startswith("_") else logging.INFO,
                           "executing %s ...", task)
                task.execute(stop)

                # Check that all targets were created.
                for target in task.targets:
                    if not target.is_file():
                        raise FileNotFoundError(f"task {task} did not create target {target}")
                    LOGGER.debug("%s created `%s`", task, target)

                # Add the result to the output queue and report success.
                output_queue.put((task, None, digest))
                delta = util.format_timedelta(datetime.now() - start)
                LOGGER.log(logging.DEBUG if task.name.startswith("_") else logging.INFO,
                           "completed %s in %s", task, delta)
            except:  # noqa: E722
                exc_info = sys.exc_info()
                delta = util.format_timedelta(datetime.now() - start)
                LOGGER.exception("failed to execute %s after %s", task, delta, exc_info=exc_info)
                stop.set()
                output_queue.put((task, sys.exc_info(), digest))

        # Put anything on the queue in case the parent is waiting.
        LOGGER.debug(f"exiting thread `{threading.current_thread().name}`")
        output_queue.put(None)

    def reset(self, *tasks: "Task") -> None:
        # TODO: add tests for resetting.
        params = [{"name": task.name} for task in tasks]
        cursor = self.connection.executemany(
            "UPDATE tasks SET digest = '__reset__' WHERE name = :name", params
        )
        self.connection.commit()
        n_reset = cursor.rowcount
        LOGGER.info("reset %d %s", n_reset, "task" if n_reset == 1 else "tasks")



from cook.__main__ import __main__
from cook.util import working_directory
from pathlib import Path
import pytest


@pytest.mark.parametrize("name, task", [
    ("hellomake", "say-hello"),
])
def test_example(name: str, task: str) -> None:
    with working_directory(f"examples/{name}"):
        # Remove the database file before testing for consistent execution.
        db = Path(".cook")
        if db.is_file():
            db.unlink()
        __main__(["exec", task])



from cook.__main__ import __main__, Formatter
import logging
from pathlib import Path
import pytest
import shutil
import sys
from typing import List


RECIPES = Path(__file__).parent / "recipes"


def test_blah_recipe_run(tmp_wd: Path) -> None:
    __main__(["--recipe", str(RECIPES / "blah.py"), "exec", "run"])


@pytest.mark.parametrize("patterns, expected", [
    (["*"], ["create_source", "compile", "link", "run"]),
    (["c*"], ["create_source", "compile"]),
    (["--re", r"^\w{3}\w?$"], ["link", "run"]),
    (["run"], ["run"]),
])
def test_blah_recipe_ls(patterns: str, expected: List[str], capsys: pytest.CaptureFixture) -> None:
    __main__(["--recipe", str(RECIPES / "blah.py"), "ls", *patterns])
    out, _ = capsys.readouterr()
    for task in expected:
        assert f"<task `{task}` @ " in pytest.shared.strip_colors(out)


def test_blah_recipe_info(tmp_wd: Path, capsys: pytest.CaptureFixture) -> None:
    __main__(["--recipe", str(RECIPES / "blah.py"), "info", "link"])
    stdout, _ = capsys.readouterr()
    assert "status: stale" in pytest.shared.strip_colors(stdout)

    __main__(["--recipe", str(RECIPES / "blah.py"), "exec", "link"])
    __main__(["--recipe", str(RECIPES / "blah.py"), "info", "link"])
    stdout, _ = capsys.readouterr()
    assert "status: current" in pytest.shared.strip_colors(stdout)

    __main__(["--recipe", str(RECIPES / "blah.py"), "info", "run"])
    stdout, _ = capsys.readouterr()
    assert "targets: -" in pytest.shared.strip_colors(stdout)

    # Check filtering based on stale/current status.
    __main__(["--recipe", str(RECIPES / "blah.py"), "info", "--stale"])
    stdout, _ = capsys.readouterr()
    assert "status: current" not in pytest.shared.strip_colors(stdout)

    __main__(["--recipe", str(RECIPES / "blah.py"), "info", "--current"])
    stdout, _ = capsys.readouterr()
    assert "status: stale" not in pytest.shared.strip_colors(stdout)

    __main__(["--recipe", str(RECIPES / "blah.py"), "info"])
    stdout, _ = capsys.readouterr()
    stdout = pytest.shared.strip_colors(stdout)
    assert "status: stale" in stdout and "status: current" in stdout

    # Check only one can be given.
    with pytest.raises(ValueError, match="may be given at the same time"):
        __main__(["--recipe", str(RECIPES / "blah.py"), "info", "--current", "--stale"])


def test_blah_recipe_reset(tmp_wd: Path) -> None:
    __main__(["--recipe", str(RECIPES / "blah.py"), "reset", "link"])


def test_simple_dag_run(tmp_wd: Path) -> None:
    __main__(["--recipe", str(RECIPES / "simple_dag.py"), "exec", "3-1"])


@pytest.mark.parametrize("patterns", [
    ["foo"],
    ["foo", "bar"],
    ["foo", "bar", "baz"],
    ["*hidden*"],
])
def test_simple_dag_no_matching_tasks(caplog: pytest.LogCaptureFixture, patterns: List[str]) \
        -> None:
    with pytest.raises(SystemExit), caplog.at_level("WARNING"):
        __main__(["--recipe", str(RECIPES / "simple_dag.py"), "ls", *patterns])
    assert "found no tasks matching" in caplog.text


def test_module_import(tmp_wd: Path) -> None:
    recipe = tmp_wd / "my_recipe.py"
    shutil.copy(RECIPES / "simple_dag.py", recipe)
    __main__(["-m", "my_recipe", "ls"])


def test_bad_recipe(caplog: pytest.LogCaptureFixture) -> None:
    with pytest.raises(SystemExit), caplog.at_level("ERROR"):
        __main__(["--recipe", str(RECIPES / "bad.py"), "exec", "false"])
    assert "failed to execute" in caplog.text


def test_custom_formatter() -> None:
    try:
        raise ValueError("terrible error")
    except ValueError:
        exc_info = sys.exc_info()
    formatter = Formatter()
    record = logging.LogRecord("a", logging.ERROR, "b", 2, "foo", None, exc_info=exc_info)
    formatted = formatter.format(record)
    assert isinstance(formatted, str)
    assert "terrible error" in formatted


def test_terrible_recipe(caplog: pytest.LogCaptureFixture) -> None:
    with pytest.raises(SystemExit), caplog.at_level("CRITICAL"):
        __main__(["--recipe", str(RECIPES / "terrible.not-py"), "ls"])
    assert "failed to load recipe" in caplog.text



from cook.actions import CompositeAction, FunctionAction, ModuleAction, SubprocessAction
from cook.util import StopEvent, Timer
from pathlib import Path
import pytest
from subprocess import SubprocessError
import sys
import threading
import time
from unittest import mock


def test_shell_action(tmp_wd: Path) -> None:
    action = SubprocessAction("echo hello > world.txt", shell=True)
    action.execute(None)
    assert (tmp_wd / "world.txt").read_text().strip() == "hello"


def test_shell_action_timeout() -> None:
    stop = StopEvent(0.01)

    def target():
        time.sleep(1)
        stop.set()

    thread = threading.Thread(target=target)
    thread.start()

    action = SubprocessAction(["sleep", "2"])
    with Timer() as timer, pytest.raises(SubprocessError, match="SIGTERM"):
        action.execute(None, stop)

    assert 1 < timer.duration < 2
    assert not thread.is_alive()


def test_subprocess_action(tmp_wd: Path) -> None:
    action = SubprocessAction(["touch", "foo"])
    action.execute(None)
    assert (tmp_wd / "foo").is_file()


def test_bad_subprocess_action() -> None:
    action = SubprocessAction("false")
    with pytest.raises(SubprocessError):
        action.execute(None)


def test_subprocess_action_repr() -> None:
    assert repr(SubprocessAction("false")) == "SubprocessAction('false')"
    assert repr(SubprocessAction(["wait", "3"])) == "SubprocessAction('wait 3')"


def test_function_action() -> None:
    args = []

    action = FunctionAction(args.append)
    action.execute(42)

    assert args == [42]


def test_composite_action() -> None:
    args = []

    action = CompositeAction(FunctionAction(args.append), FunctionAction(args.append))
    action.execute("hello")
    assert args == ["hello", "hello"]


def test_module_action() -> None:
    action = ModuleAction([pytest, "-h"])
    action.execute(None)

    with pytest.raises(ValueError, match="shell execution"):
        ModuleAction([pytest], shell=True)

    with pytest.raises(ValueError, match="not be empty"):
        ModuleAction([])

    with pytest.raises(TypeError, match="must be a module"):
        ModuleAction([None])


def test_module_action_debug() -> None:
    with mock.patch("subprocess.Popen") as Popen:
        Popen().wait = mock.MagicMock(return_value=0)
        ModuleAction([pytest], debug=True).execute(None)
    Popen.assert_called_with([sys.executable, "-m", "pdb", "-m", "pytest"])

    with mock.patch("subprocess.Popen") as Popen:
        Popen().wait = mock.MagicMock(return_value=0)
        ModuleAction([pytest], debug=False).execute(None)
    Popen.assert_called_with([sys.executable, "-m", "pytest"])


def test_composite_digest() -> None:
    actions = [
        SubprocessAction("hello"),
        SubprocessAction(["foo", "bar"]),
    ]
    assert CompositeAction(*actions).hexdigest

    actions.append(FunctionAction(print))
    assert CompositeAction(*actions).hexdigest is None



"""
Actions
-------

Actions are performed when tasks are executed. Builtin actions include calling python functions
using :class:`.FunctionAction`, running subprocesses using :class:`.SubprocessAction`, composing
multiple actions using :class:`.CompositeAction`, and executing modules as scripts using
:class:`.ModuleAction`.

Custom actions can be implemented by inheriting from :class:`.Action` and implementing the
:meth:`~.Action.execute` method which receives a :class:`~.task.Task`. The method should execute the
action; its return value is ignored. For example, the following action waits for a specified time.

.. doctest::

    >>> from cook.actions import Action
    >>> from cook.task import Task
    >>> from time import sleep, time

    >>> class SleepAction(Action):
    ...     def __init__(self, delay: float) -> None:
    ...         self.delay = delay
    ...
    ...     def execute(self, task: Task) -> None:
    ...         start = time()
    ...         sleep(self.delay)
    ...         print(f"time: {time() - start:.3f}")

    >>> action = SleepAction(0.1)
    >>> action.execute(None)
    time: 0.1...
"""
import hashlib
import os
import shlex
import subprocess
import sys
from types import ModuleType
from typing import Callable, List, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    from .task import Task
    from .util import StopEvent


class Action:
    """
    Action to perform when a task is executed in its own thread.
    """
    def execute(self, task: "Task", stop: Optional["StopEvent"] = None) -> None:
        """
        Execute the action.
        """
        raise NotImplementedError

    @property
    def hexdigest(self) -> Optional[str]:
        """
        Optional digest to check if an action changed.
        """
        return None


class FunctionAction(Action):
    """
    Action wrapping a python callable.

    Args:
        func: Function to call which must accept a :class:`~.task.Task` as its first argument.
        *args: Additional positional arguments.
        **kwargs: Keyword arguments.
    """
    def __init__(self, func: Callable, *args, **kwargs) -> None:
        super().__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def execute(self, task: "Task", stop: Optional["StopEvent"] = None) -> None:
        self.func(task, *self.args, **self.kwargs)


class SubprocessAction(Action):
    """
    Run a subprocess.

    Args:
        *args: Positional arguments for :class:`subprocess.Popen`.
        **kwargs: Keyword arguments for :class:`subprocess.Popen`.

    Example:

        .. doctest::

            >>> from cook.actions import SubprocessAction
            >>> from pathlib import Path

            >>> action = SubprocessAction(["touch", "hello.txt"])
            >>> action.execute(None)
            >>> Path("hello.txt").is_file()
            True
    """
    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs

    def execute(self, task: "Task", stop: Optional["StopEvent"] = None) -> None:
        # Repeatedly wait for the process to complete, checking the stop event after each poll.
        interval = stop.interval if stop else None
        process = subprocess.Popen(*self.args, **self.kwargs)
        while True:
            try:
                returncode = process.wait(interval)
                if returncode:
                    raise subprocess.CalledProcessError(returncode, process.args)
                return
            except subprocess.TimeoutExpired:
                if stop.is_set():
                    break

        # Clean up the process by trying to terminate it and then killing it.
        for method in [process.terminate, process.kill]:
            method()
            try:
                returncode = process.wait(max(interval, 3))
                if returncode:
                    raise subprocess.CalledProcessError(returncode, process.args)
                # The process managed to exit gracefully after the main loop. This is unlikely.
                return  # pragma: no cover
            except subprocess.TimeoutExpired:  # pragma: no cover
                pass

        # We couldn't kill the process. Also very unlikely.
        raise subprocess.SubprocessError(f"failed to shut down {process}")  # pragma: no cover

    @property
    def hexdigest(self) -> str:
        hasher = hashlib.sha1()
        args, = self.args
        if isinstance(args, str):
            hasher.update(args.encode())
        else:
            for arg in args:
                hasher.update(arg.encode())
        return hasher.hexdigest()

    def __repr__(self) -> str:
        args, *_ = self.args
        if not isinstance(args, str):
            args = ' '.join(map(shlex.quote, args))
        return f"{self.__class__.__name__}({repr(args)})"


class CompositeAction(Action):
    """
    Execute multiple actions sequentially.

    Args:
        *actions: Actions to execute.
    """
    def __init__(self, *actions: Action) -> None:
        self.actions = actions

    def execute(self, task: "Task", stop: Optional["StopEvent"] = None) -> None:
        for action in self.actions:
            action.execute(task, stop)

    @property
    def hexdigest(self) -> Optional[str]:
        parts = []
        for action in self.actions:
            hexdigest = action.hexdigest
            if hexdigest is None:
                return None
            parts.append(hexdigest)
        return "".join(parts)


class ModuleAction(SubprocessAction):
    """
    Execute a module as a script.

    Args:
        args: List comprising the module to execute as the first element and arguments for the
            module as subsequent elements.
        debug: Run the module using `pdb` (defaults to the :code:`COOK_DEBUG` environment variable
            being set).
        **kwargs: Keyword arguments for :class:`subprocess.Popen`.
    """
    def __init__(self, args: List, debug: Optional[bool] = None, **kwargs) -> None:
        if kwargs.get("shell"):
            raise ValueError("shell execution is not supported by `ModuleAction`")
        if not args:
            raise ValueError("`args` must not be empty")
        module: ModuleType
        module, *args = args
        if not isinstance(module, ModuleType):
            raise TypeError("first element of `args` must be a module")

        # Assemble the arguments.
        args_ = [sys.executable, "-m"]
        debug = "COOK_DEBUG" in os.environ if debug is None else debug
        if debug:
            args_.extend(["pdb", "-m"])
        args_.extend([module.__name__, *map(str, args)])
        super().__init__(args_, **kwargs)



from cook import util
from datetime import datetime, timedelta
import hashlib
from pathlib import Path


def test_evaluate_digest(tmp_wd: Path) -> None:
    fn = tmp_wd / "foo.txt"
    fn.write_text("bar")
    assert util.evaluate_hexdigest(fn) == hashlib.sha1(b"bar").hexdigest()


def test_format_timedelta() -> None:
    assert util.format_timedelta(timedelta(1, 13, 17, 28, 40, 3, 8)) == "57 days, 3:40:13"
    assert util.format_timedelta(timedelta(microseconds=999)) == "0:00:00.000999"


def test_format_datetime() -> None:
    assert util.format_datetime(datetime(2023, 7, 25, 13, 7, 9, 777)) == "2023-07-25 13:07:09"



from setuptools import find_packages, setup


with open("README.rst") as fp:
    long_description = fp.read()
long_description = long_description \
    .replace(":func:", ":code:") \
    .replace(".. doctest::", ".. code-block::\n") \
    .replace(".. shtest::", ".. code-block::\n") \
    .replace(":class:", ":code:") \
    .replace(".. toctree::", "..") \
    .replace(".. sh::", "..")


setup(
    name="cook-build",
    version="0.5.1",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    python_requires=">=3.8",
    install_requires=[
        "colorama",
        "networkx",
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "cook = cook.__main__:__main__",
        ],
    },
)



from __future__ import annotations
import contextlib
from datetime import datetime, timedelta
import hashlib
import inspect
import os
from pathlib import Path
import threading
from time import time
from typing import Tuple, TYPE_CHECKING, Union


if TYPE_CHECKING:
    from .task import Task


PathOrStr = Union[Path, str]


def evaluate_digest(path: PathOrStr, size=2 ** 16, hasher: str = "sha1") -> bytes:
    hasher = hashlib.new(hasher)
    path = Path(path)
    with path.open("rb") as fp:
        while chunk := fp.read(size):
            hasher.update(chunk)
    return hasher.digest()


def evaluate_hexdigest(path: PathOrStr, size=2 ** 16, hasher: str = "sha1") -> str:
    return evaluate_digest(path, size, hasher).hex()


class Timer:
    def __init__(self):
        self.start = None

    def __enter__(self) -> Timer:
        self.start = time()
        return self

    def __exit__(self, *_) -> None:
        self.end = time()

    @property
    def duration(self):
        return self.end - self.start


class CookError(Exception):
    pass


class FailedTaskError(Exception):
    def __init__(self, *args: object, task: "Task") -> None:
        super().__init__(*args)
        self.task = task


@contextlib.contextmanager
def working_directory(path: PathOrStr) -> Path:
    path = Path(path)
    original = Path.cwd()
    try:
        os.chdir(path)
        yield path
    finally:
        os.chdir(original)


def get_location() -> Tuple[str, int]:
    """
    Get the first location in the call stack which is not part of the Cook package.

    Returns:
        Location as a tuple :code:`(filename, lineno)`.
    """
    frame = inspect.currentframe()
    while frame.f_globals.get("__name__", "<unknown>").startswith("cook"):
        frame = frame.f_back
    return Path(frame.f_code.co_filename).resolve(), frame.f_lineno


class StopEvent(threading.Event):
    """
    Event used for stopping execution with a polling interval.
    """
    def __init__(self, interval: float = 1) -> None:
        super().__init__()
        self.interval = interval


def format_timedelta(delta: timedelta) -> str:
    """
    Format a time difference.
    """
    if delta.total_seconds() < 1:
        return str(delta)
    return str(delta).rsplit(".", 2)[0]


def format_datetime(dt: datetime) -> str:
    """
    Format a date-time.
    """
    return str(dt).rsplit(".", 2)[0]



from cook import Controller, Manager, Task
from cook.actions import Action, CompositeAction, FunctionAction, SubprocessAction
from cook.contexts import normalize_action, normalize_dependencies
from cook.controller import QUERIES
from cook.util import FailedTaskError, StopEvent, Timer
from datetime import datetime
from pathlib import Path
import pytest
import shutil
from sqlite3 import Connection
import threading
import time
from unittest.mock import patch


def touch(task: Task) -> None:
    for target in task.targets:
        target.write_text(target.name)


def test_controller_empty_task(m: Manager, conn: Connection) -> None:
    task = m.create_task("foo")
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task)
    assert c.resolve_stale_tasks() == {task}


def test_controller_missing_target(m: Manager, conn: Connection) -> None:
    task = m.create_task("foo", targets=["bar"])
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task)
    assert c.resolve_stale_tasks() == {task}


def test_controller_simple_file_deps(m: Manager, conn: Connection, tmp_wd: Path) -> None:
    for path in ["input.txt", "output.txt"]:
        Path(path).write_text(path)
    with normalize_dependencies():
        task = m.create_task("foo", dependencies=["input.txt"], targets=["output.txt"])
    c = Controller(m.resolve_dependencies(), conn)

    # No entry in the database.
    assert c.is_stale(task)
    assert c.resolve_stale_tasks() == {task}

    # Up to date entry in the database.
    params = {
        "name": "foo",
        "digest": "80d4129af3d5366c3fcd26c498e143d9a199f7c4",
        "last_completed": None,
    }
    conn.execute(QUERIES["upsert_task_completed"], params)
    c = Controller(m.resolve_dependencies(), conn)
    assert not c.is_stale(task)

    # Wrong digest in the database.
    params = {"name": "foo", "digest": "-", "last_completed": None}
    conn.execute(QUERIES["upsert_task_completed"], params)
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task)


def test_controller_missing_input(m: Manager, conn: Connection) -> None:
    with normalize_dependencies():
        m.create_task("input", targets=["input.txt"], action=FunctionAction(touch))
        output = m.create_task("output", targets=["output.txt"], action=FunctionAction(touch),
                               dependencies=["input.txt"])

    # Create the output.
    Path("output.txt").write_text("output.txt")
    params = {
        "name": "output",
        "digest": "80d4129af3d5366c3fcd26c498e143d9a199f7c4",
        "last_completed": None,
    }
    conn.execute(QUERIES["upsert_task_completed"], params)
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(output) is True


def test_controller(m: Manager, conn: Connection) -> None:
    for filename in ["input1.txt", "input2.txt", "intermediate.txt", "output1.txt"]:
        Path(filename).write_text(filename)

    with normalize_dependencies():
        intermediate = m.create_task("intermediate", dependencies=["input1.txt", "input2.txt"],
                                     targets=["intermediate.txt"], action=FunctionAction(touch))
        output1 = m.create_task("output1", dependencies=["intermediate.txt"],
                                targets=["output1.txt"], action=FunctionAction(touch))
        output2 = m.create_task("output2", targets=["output2.txt"], action=FunctionAction(touch),
                                dependencies=["intermediate.txt", "input2.txt", "output1.txt"])
        special = m.create_task("special", dependencies=["intermediate.txt"])

    # We should get back all tasks at the beginning.
    c = Controller(m.resolve_dependencies(), conn)
    assert c.resolve_stale_tasks() == {intermediate, output1, output2, special}

    # Make sure we don't get any tasks that are upstream from what we request.
    c = Controller(m.resolve_dependencies(), conn)
    assert c.resolve_stale_tasks([output1]) == {intermediate, output1}

    # Execute tasks and check that they are no longer stale.
    c = Controller(m.resolve_dependencies(), conn)
    c.execute(output1)
    assert not c.resolve_stale_tasks([output1])

    # But the other ones are still stale.
    c = Controller(m.resolve_dependencies(), conn)
    assert c.resolve_stale_tasks() == {output2, special}

    # Execute the second output. The special task without outputs never disappears.
    c = Controller(m.resolve_dependencies(), conn)
    c.execute(output2)
    assert c.resolve_stale_tasks() == {special}


def test_target_not_created(m: Manager, conn: Connection) -> None:
    task = m.create_task("nothing", targets=["missing"])
    c = Controller(m.resolve_dependencies(), conn)
    with pytest.raises(FailedTaskError, match="did not create"):
        c.execute(task)


def test_failing_task(m: Manager, conn: Connection) -> None:
    def raise_exception(_) -> None:
        raise RuntimeError

    task = m.create_task("nothing", action=FunctionAction(raise_exception))
    c = Controller(m.resolve_dependencies(), conn)
    with pytest.raises(FailedTaskError):
        c.execute(task)
    last_failed, = conn.execute("SELECT last_failed FROM tasks WHERE name = 'nothing'").fetchone()
    assert (datetime.now() - last_failed).total_seconds() < 1


def test_concurrency(m: Manager, conn: Connection) -> None:
    delay = 0.2
    num_tasks = 4

    tasks = [
        m.create_task(
            str(i), action=SubprocessAction(f"sleep {delay} && touch {i}.txt", shell=True),
            targets=[f"{i}.txt"]
        ) for i in range(num_tasks)
    ]
    task = m.create_task("result", dependencies=[task.targets[0] for task in tasks])

    c = Controller(m.resolve_dependencies(), conn)
    with Timer() as timer:
        c.execute(task)
    assert timer.duration > num_tasks * delay

    c = Controller(m.resolve_dependencies(), conn)
    with Timer() as timer:
        c.execute(task, num_concurrent=num_tasks)
    assert timer.duration < 2 * delay


def test_digest_cache(m: Manager, conn: Connection, tmp_wd: Path) -> None:
    c = Controller(m.resolve_dependencies(), conn)
    shutil.copy(__file__, tmp_wd / "foo")
    with patch("cook.util.evaluate_digest", return_value=b"aaaa") as evaluate_digest:
        c._evaluate_path_hexdigest("foo")
        c._evaluate_path_hexdigest("foo")
    evaluate_digest.assert_called_once()


def test_skip_if_no_stale_tasks(m: Manager, conn: Connection, tmp_wd: Path) -> None:
    c = Controller(m, conn)
    c.execute([])


def test_tasks_are_executed(m: Manager, conn: Connection, tmp_wd: Path) -> None:
    with normalize_dependencies():
        m.create_task("base", targets=["base.txt"], action=FunctionAction(touch))
        task1 = m.create_task("task1", targets=["task1.txt"], action=FunctionAction(touch),
                              dependencies=["base.txt"])
        task2 = m.create_task("task2", targets=["task2.txt"], action=FunctionAction(touch),
                              dependencies=["base.txt"])
    # Execute the first task.
    c = Controller(m.resolve_dependencies(), conn)
    c.execute(task1)

    # Verify the second is still stale.
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task2)

    with open("task2.txt", "w"):
        pass

    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task2)


def test_is_stale_called_once(m: Manager, conn: Connection) -> None:
    with normalize_dependencies():
        m.create_task("base", targets=["base.txt"], action=FunctionAction(touch))
        m.create_task("task1", targets=["task1.txt"], action=FunctionAction(touch),
                      dependencies=["base.txt"])
        m.create_task("task2", targets=["task2.txt"], action=FunctionAction(touch),
                      dependencies=["base.txt"])

    with patch("cook.controller.Controller._is_self_stale", return_value=False) as _is_self_stale:
        c = Controller(m.resolve_dependencies(), conn)
        [c.is_stale(list(c.dependencies)) for _ in range(7)]

    assert _is_self_stale.call_count == 3


def test_stop_long_running_subprocess(m: Manager, conn: Connection) -> None:
    task = m.create_task("sleep", action=SubprocessAction(["sleep", "60"]))
    c = Controller(m.resolve_dependencies(), conn)

    # Set the stop interval in one second.
    stop = StopEvent(0.01)

    def _target():
        time.sleep(1)
        stop.set()
    thread = threading.Thread(target=_target)
    thread.start()

    with patch("cook.util.StopEvent", return_value=stop), Timer() as timer, \
            pytest.raises(FailedTaskError, match="SIGTERM: 15"):
        c.execute(task)

    assert 1 < timer.duration and timer.duration < 2
    assert not thread.is_alive()


def test_set_stop_between_tasks(m: Manager, conn: Connection) -> None:
    calls = []

    def _action(task):
        calls.append(task)
        time.sleep(0.5)

    task1 = m.create_task("sleep1", action=FunctionAction(_action))
    task2 = m.create_task("sleep2", action=FunctionAction(_action))
    c = Controller(m.resolve_dependencies(), conn)

    # Set the stop interval shortly after dispatching the tasks.
    stop = StopEvent(10)

    def _target():
        time.sleep(0.3)
        stop.set()
    thread = threading.Thread(target=_target)
    thread.start()

    with patch("cook.util.StopEvent", return_value=stop), Timer() as timer:
        c.execute([task1, task2])

    assert 0.5 < timer.duration and timer.duration < 1
    assert not thread.is_alive()
    assert len(calls) == 1


def test_no_start_after_failure(m: Manager, conn: Connection) -> None:
    calls = []

    def _action(task):
        calls.append(task)
        raise ValueError

    task1 = m.create_task("sleep1", action=FunctionAction(_action))
    task2 = m.create_task("sleep2", action=FunctionAction(_action))
    c = Controller(m.resolve_dependencies(), conn)

    with pytest.raises(FailedTaskError):
        c.execute([task1, task2])

    assert len(calls) == 1


def test_digest_race(m: Manager, conn: Connection) -> None:
    with normalize_dependencies(), normalize_action():
        task1 = m.create_task("task1", action="echo foo > bar.txt", targets=["bar.txt"])
        task2 = m.create_task("task2", action="sleep 1 && echo hello > world.txt",
                              targets=["world.txt"], dependencies=["bar.txt"])

    # Execute and ...
    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task2)
    c.execute(task2)

    # ... ensure the task is no longer stale.
    c = Controller(m.resolve_dependencies(), conn)
    assert not c.is_stale(task2)

    # Modify the file and check the task is stale.
    c = Controller(m.resolve_dependencies(), conn)
    task1.targets[0].write_text("buzz")
    assert c.is_stale(task2)

    # Execute again but modify the dependency during execution.
    c = Controller(m.resolve_dependencies(), conn)

    def _target():
        time.sleep(0.5)
        task1.targets[0].write_text("fizz")
    thread = threading.Thread(target=_target)

    thread.start()
    c.execute(task2)
    assert not thread.is_alive()

    c = Controller(m.resolve_dependencies(), conn)
    assert c.is_stale(task2)


@pytest.mark.parametrize("action1, action2", [
    (
        SubprocessAction("echo hello > b.txt", shell=True),
        SubprocessAction("echo fizz > b.txt", shell=True)),
    (
        CompositeAction(SubprocessAction("echo a > b.txt", shell=True), SubprocessAction("true")),
        CompositeAction(SubprocessAction("echo a > b.txt", shell=True), SubprocessAction("false"))
    )
])
def test_action_digest(m: Manager, conn: Connection, action1: Action, action2: Action) -> None:
    with normalize_dependencies():
        task = m.create_task("task", action=action1, targets=["b.txt"])

    # Execute and check stale status.
    Controller(m.resolve_dependencies(), conn).execute(task)
    assert not Controller(m.resolve_dependencies(), conn).is_stale(task)

    # Modify the task and check stale status.
    task.action = action2
    assert Controller(m.resolve_dependencies(), conn).is_stale(task)


# TODO: add tests to verify what happens when tasks are cancelled, e.g., by `KeyboardInterrupt`.
